import { Injectable, Logger } from '@nestjs/common';
import { ConfigService } from '@nestjs/config';
import { createClient, SupabaseClient } from '@supabase/supabase-js';

@Injectable()
export class SupabaseService {
  private supabase: SupabaseClient;
  private supabase_url: string;
  private supabase_key: string;
  private readonly logger = new Logger(SupabaseService.name);

  constructor(private configService: ConfigService) {
    // Initialize properties in the constructor
    this.supabase_url = this.configService.get<string>('SUPABASE_URL') || '';
    this.supabase_key = this.configService.get<string>('SUPABASE_KEY') || '';

    this.supabase = createClient(this.supabase_url, this.supabase_key);
  }

  async insertShared(metadata: any) {
    const { error } = await this.supabase.from('shared').insert(metadata);
    if (error) {
      this.logger.error(
        `Failed to insert shared: ${error.message}`,
        error.details,
      );
      throw new Error(`Failed to insert shared: ${error.message}`);
    }
  }

  async getSharedById(id: string) {
    const { data, error } = await this.supabase
      .from('shared')
      .select('*')
      .eq('id', id)
      .single();
    if (error) {
      this.logger.error(
        `Failed to fetch shared with ID ${id}: ${error.message}`,
        error.details,
      );
      throw new Error(`Failed to fetch shared: ${error.message}`);
    }
    return data;
  }

  async getSharedBySlug(type: string, slug: string) {
    const { data, error } = await this.supabase
      .from('shared')
      .select('*')
      .eq('type', type)
      .eq('slug', slug)
      .single();
    if (error) {
      this.logger.error(
        `Failed to fetch shared with slug ${slug}: ${error.message}`,
        error.details,
      );
      return null;
    }
    return data;
  }

  async getSharedByName(name: string) {
    const { data, error } = await this.supabase
      .from('shared')
      .select('*')
      .eq('name', name)
      .single();
    if (error && error.code !== 'PGRST116') {
      // Handle "no rows found" separately
      this.logger.error(
        `Failed to fetch shared with name ${name}: ${error.message}`,
        error.details,
      );
      throw new Error(`Failed to fetch shared: ${error.message}`);
    }
    return data;
  }

  async updateShared(slug: string, metadata: any) {
    const { error } = await this.supabase
      .from('shared')
      .update(metadata)
      .eq('slug', slug);
    if (error) {
      this.logger.error(
        `Failed to update shared with slug ${slug}: ${error.message}`,
        error.details,
      );
      throw new Error(`Failed to update shared: ${error.message}`);
    }
  }

  async deleteShared(slug: string) {
    const { error } = await this.supabase
      .from('shared')
      .delete()
      .eq('slug', slug);
    if (error) {
      this.logger.error(
        `Failed to delete shared with slug ${slug}: ${error.message}`,
        error.details,
      );
      throw new Error(`Failed to delete shared: ${error.message}`);
    }
  }

  async incrementVisits(slug: string) {
    const { error } = await this.supabase.rpc('increment_visits', {
      slug_param: slug,
    });

    if (error) {
      this.logger.error(
        `Failed to increment visits for shared with slug ${slug}: ${error.message}`,
        error.details,
      );
      throw new Error(`Failed to increment visits: ${error.message}`);
    }
  }

  async updateLevel(slug: string, level: number) {
    const { error } = await this.supabase
      .from('shared')
      .update({ level })
      .eq('slug', slug);
    if (error) {
      this.logger.error(
        `Failed to update level for shared with slug ${slug}: ${error.message}`,
        error.details,
      );
      throw new Error(`Failed to update shared level: ${error.message}`);
    }
  }

  async countSharedByToken(token: string): Promise<number> {
    const { data, error, count } = await this.supabase
      .from('shared') // Nombre de tu tabla en Supabase
      .select('*', { count: 'exact' })
      .eq('token', token);

    if (error) {
      console.error('Error counting shared by token:', error);
      throw new Error('Failed to count shared by token');
    }

    return count || 0;
  }
}
